{ File   : PAKExt.dpr
  Author : Deniz Oezmen
  Created: 2010-11-16
  Changed: 2010-11-16

  Gray Matter PAK extractor.
}
program PAKExt;

{$APPTYPE CONSOLE}

{$R '..\res\version.res' '..\res\version.rc'}

uses
  SysUtils,
  ZLib;

type
  TParamRec = record
    FileSpec: AnsiString;
  end;

  TPAKHeader = packed record
    DirUncompSize: Longint;
    DirCompSize  : Longint;
  end;

  TString = record
    Len    : Longint;
    Content: AnsiString;
  end;

  TDirEntry = record
    Name      : TString;
    Offset    : Longint;
    CompSize  : Longint;
    UncompSize: Longint;
  end;

  TDirEntries = array of TDirEntry;

  TBuffer = array of Byte;

var
  ParamRec : TParamRec;

function OpenInFile(var InFile: file; InName: string): Boolean;
begin
  FileMode := fmShareDenyWrite;
  AssignFile(InFile, InName);

  {$I-}
  Reset(InFile, 1);
  {$I+}
  Result := (IOResult = 0)
end;

function OpenOutFile(var OutFile: file; OutName: string): Boolean;
var
  Dir: AnsiString;
begin
  // first create the directory portion of the output file ...
  Dir := ExtractFileDir(ExpandFileName(OutName));
  ForceDirectories(Dir);

  // ... then open the output file
  FileMode := fmOpenWrite;
  AssignFile(OutFile, OutName);

  {$I-}
  Rewrite(OutFile, 1);
  {$I+}
  Result := (IOResult = 0)
end;

function ReadHeader(var PAKFile: file; out PAKHeader: TPAKHeader): Boolean;
var
  BytesRead : Longint;
begin
  BlockRead(PAKFile, PAKHeader, SizeOf(PAKHeader), BytesRead);

  Result := (SizeOf(PAKHeader) = BytesRead)
end;

function Decompress(const InBuffer: TBuffer; var OutBuffer: TBuffer): Boolean;
var
  DecompSize: Longint;
  TempBuffer: Pointer;
begin
  // standard ZLib decompression
  DecompressBuf(@InBuffer[0], Length(InBuffer), Length(OutBuffer), TempBuffer,
    DecompSize);

  Result := (DecompSize = Length(OutBuffer));

  if Result then
    Move(TempBuffer^, OutBuffer[0], DecompSize);

  FreeMem(TempBuffer, DecompSize)
end;

function GetIntFromBuffer(const Buffer: TBuffer; var Index: Longint;
  out Data: Longint): Boolean;
begin
  Result := (Index >= 0) and (Index + SizeOf(Data) <= Length(Buffer));
  if not Result then
    Exit;

  Move(Buffer[Index], Data, SizeOf(Data));
  Inc(Index, SizeOf(Data))
end;

function GetStringFromBuffer(const Buffer: TBuffer; var Index: Longint;
  out Data: TString): Boolean;
begin
  Result := GetIntFromBuffer(Buffer, Index, Data.Len);
  if not Result then
    Exit;

  Result := (Index + Data.Len <= Length(Buffer));
  if not Result then
    Exit;

  SetLength(Data.Content, Data.Len);

  if Data.Len > 0 then
    Move(Buffer[Index], Data.Content[1], Data.Len);
  Inc(Index, Data.Len)
end;

function ProcessDir(const DirBuffer: TBuffer; Dir: AnsiString;
  var Index: Longint; var DirEntries: TDirEntries): Boolean;
var
  i,
  DirIdx,
  NumSubDirs,
  NumFiles  : Longint;
  SubDir    : TString;
begin
  Result := False;

  // read subdirectories
  if not GetIntFromBuffer(DirBuffer, Index, NumSubDirs) then
    Exit;

  for i := 1 to NumSubDirs do
  begin
    if not GetStringFromBuffer(DirBuffer, Index, SubDir) then
      Exit;

    if not ProcessDir(DirBuffer, Dir + SubDir.Content + '\', Index,
      DirEntries) then
      Exit
  end;

  // read files within this directory
  if not GetIntFromBuffer(DirBuffer, Index, NumFiles) then
    Exit;

  DirIdx := Length(DirEntries);
  SetLength(DirEntries, Length(DirEntries) + NumFiles);

  for i := DirIdx to DirIdx + NumFiles - 1 do
  begin
    if not GetStringFromBuffer(DirBuffer, Index, DirEntries[i].Name) then
      Exit;
    if not GetIntFromBuffer(DirBuffer, Index, DirEntries[i].Offset) then
      Exit;
    if not GetIntFromBuffer(DirBuffer, Index, DirEntries[i].CompSize) then
      Exit;
    if not GetIntFromBuffer(DirBuffer, Index, DirEntries[i].UncompSize) then
      Exit;

    // add directory information to file entry
    DirEntries[i].Name.Content := Dir + DirEntries[i].Name.Content
  end;

  Result := True
end;

procedure ExtractFiles(var PAKFile: file; const PAKHeader: TPAKHeader);
var
  i,
  Idx       : Longint;
  BaseOfs   : Longint;
  CompBuf,
  UncompBuf : TBuffer;
  DirEntries: TDirEntries;
  OutFile   : file;
begin
  SetLength(CompBuf, PAKHeader.DirCompSize - SizeOf(PAKHeader));

  {$I-}
  BlockRead(PAKFile, CompBuf[0], Length(CompBuf));
  {$I+}
  if IOResult <> 0 then
  begin
    WriteLn('  error reading structures');
    SetLength(CompBuf, 0);

    Exit
  end;

  BaseOfs := FilePos(PAKFile);

  SetLength(UncompBuf, PAKHeader.DirUncompSize);

  if Decompress(CompBuf, UncompBuf) then
  begin
    // read and prepare file information
    Idx := 0;
    SetLength(DirEntries, 0);
    if ProcessDir(UncompBuf, '', Idx, DirEntries) then
      for i := 0 to High(DirEntries) do
      begin
        WriteLn('  ' + DirEntries[i].Name.Content);

        // read compressed file data
        SetLength(CompBuf, DirEntries[i].CompSize);
        {$I-}
        Seek(PAKFile, BaseOfs + DirEntries[i].Offset);
        BlockRead(PAKFile, CompBuf[0], Length(CompBuf));
        {$I+}
        if IOResult <> 0 then
        begin
          WriteLn('    read error');
          Continue
        end;

        // decompress
        SetLength(UncompBuf, DirEntries[i].UncompSize);
        if not Decompress(CompBuf, UncompBuf) then
        begin
          WriteLn('    decompression error');
          Continue
        end;

        // output decompressed file
        if OpenOutFile(OutFile, DirEntries[i].Name.Content) then
        begin
          {$I-}
          BlockWrite(OutFile, UncompBuf[0], Length(UncompBuf));
          {$I+}
          if IOResult <> 0 then
            WriteLn('    write error');
           CloseFile(OutFile)
        end
        else
          WriteLn('    error opening output')
      end
    else
      WriteLn('  error reading directory entries')
  end
  else
    WriteLn('  error decompressing structures');

  SetLength(CompBuf, 0);
  SetLength(UncompBuf, 0);
  SetLength(DirEntries, 0)
end;

procedure ScanFiles;
var
  FileName,
  Dir,
  FullName : string;
  PAKFile  : file;
  PAKHeader: TPAKHeader;
  SearchRec: TSearchRec;
begin
  // get the directory information
  Dir := ExtractFilePath(ParamRec.FileSpec);

  // search for all files matching the file specification
  if FindFirst(ParamRec.FileSpec, faAnyFile - faDirectory,
    SearchRec) = 0 then
  begin
    repeat
      FileName := SearchRec.Name;
      FullName := Dir + FileName;

      WriteLn(FullName);

      if OpenInFile(PAKFile, FullName) then
      begin
        if ReadHeader(PAKFile, PAKHeader) then
          ExtractFiles(PAKFile, PAKHeader)
        else
          WriteLn('  not a valid PAK file')
      end
      else
        WriteLn('  not accessible')
    until FindNext(SearchRec) <> 0
  end
  else
    WriteLn(Format('No matches for "%s"', [ParamRec.FileSpec]));

  SysUtils.FindClose(SearchRec)
end;

function ProcessParams: Boolean;
var
  i: Longint;
begin
  Result := False;

  if ParamCount = 0 then
    Exit;

  with ParamRec do
  begin
    FileSpec := '';

    for i := 1 to ParamCount do
    begin
      if FileSpec <> '' then
        Exit
      else
        FileSpec := ParamStr(i)
    end;

    Result := (FileSpec <> '')
  end
end;

begin
  try
    if ProcessParams then
      ScanFiles
    else
      WriteLn('Usage: PAKExt <PAKFiles>')
  except
    on FatalError: Exception do
      WriteLn('FATAL ERROR: ' + FatalError.Message)
  end
end.

